from typing import List, Union

from zeno.classes.base import CamelModel


class Points2D(CamelModel):
    x: List[float]
    y: List[float]
    color: List[int]
    opacity: List[int]
    ids: List[str]
    domain: List[Union[int, float, str]]
    # "nominal" or "continuous" or "boolean"
    dataType: str



"""The `zeno init` command, which sets up a new Zeno project."""

import os

import requests
from InquirerPy.resolver import prompt
from InquirerPy.validator import PathValidator

from zeno.util import VIEW_MAP_URL, VIEWS_MAP_JSON

TOML_TEMPLATE = """functions = "./"
view = "{}"
metadata = "{}"
data_path = "{}"
data_column = "{}"
{}
models = [] # Add paths to model checkpoints, passed into @model
"""


def setup_zeno():
    print(
        "Welcome to \033[1mZeno\033[0m! The following questions will guide "
        + "you through setting up your project.\n"
    )

    views_res = requests.get(VIEW_MAP_URL + VIEWS_MAP_JSON)
    views = list(views_res.json().keys())

    questions = [
        {
            "type": "input",
            "message": "Name your project/folder:",
            "name": "name",
            "validate": lambda r: len(r) > 0,
            "invalid_message": "Input cannot be empty.",
        },
        {
            "type": "list",
            "message": "What type of model are you evaluating?",
            "name": "view",
            "choices": views,
        },
        {
            "type": "filepath",
            "message": "Path to a metadata file (CSV or Parquet):",
            "name": "metadata",
            "validate": PathValidator(is_file=True, message="Input is not a file"),
            "only_files": True,
        },
        {
            "type": "filepath",
            "message": "Folder or URL with data (e.g. image folder or S3 endpoint):",
            "name": "data_path",
        },
        {
            "type": "input",
            "message": "Name of the column in the metadata file"
            + " with relative file paths:",
            "name": "data_column",
        },
        {
            "type": "input",
            "message": "Name of the column in the metadata file"
            + " with ground truth labels (OPTIONAL):",
            "name": "label_column",
        },
        {
            "type": "list",
            "message": "Which library are you using?",
            "name": "model_type",
            "choices": ["None", "PyTorch", "TensorFlow", "Keras", "HuggingFace"],
        },
        {"type": "confirm", "message": "Confirm?", "name": "confirm"},
    ]

    result = prompt(questions)
    print(
        "\nCongrats! A new folder, {}, has been created with a ".format(result["name"])
        + "zeno.toml file inside. Run Zeno with the following:"
    )
    print(("\n\t cd {} \n\t zeno zeno.toml").format(result["name"]))

    print("\n To set up your model, follow the instructions in model.py")

    # Create new directory with zeno.toml file
    os.mkdir(str(result["name"]))

    # Update paths to be relative to new directory if relative to start with.
    if os.path.exists(str(result["data_path"])) and not os.path.isabs(str("data_path")):
        result["data_path"] = os.path.relpath(
            str(result["data_path"]), "./" + str(result["name"])
        )
    if not os.path.isabs(str("metadata")):
        result["metadata"] = os.path.relpath(
            str(result["metadata"]), "./" + str(result["name"])
        )

    os.chdir(str(result["name"]))

    with open("zeno.toml", "w") as f:
        f.write(
            TOML_TEMPLATE.format(
                result["view"],
                result["metadata"],
                result["data_path"],
                result["data_column"],
                'label_column = "' + str(result["label_column"]) + '"'
                if result["label_column"]
                else "",
            )
        )
    # Create model.py file



"""Functions for parsing filter predicates and filtering dataframes"""
from typing import List, Optional, Union

import pandas as pd
from pandas import DataFrame

from zeno.classes.base import MetadataType, ZenoColumn
from zeno.classes.metadata import HistogramBucket

from zeno.classes.slice import FilterIds, FilterPredicate, FilterPredicateGroup


def get_filter_string(filter: Union[FilterPredicateGroup, FilterPredicate]):
    if isinstance(filter, FilterPredicateGroup):
        if len(filter.predicates) > 0:
            filt = "("
            for i, pred in enumerate(filter.predicates):
                filt = filt + get_filter_string(pred)
                if i == len(filter.predicates) - 1 and (
                    filt[-1] == "&" or filt[-1] == "|"
                ):
                    filt = filt[0:-1]
            return filt + ")" + (filter.join if filter.join else "")
    else:
        if filter.operation == "match":
            return "(`{}`.str.match('{}', na=False)) {}".format(
                filter.column, filter.value, filter.join
            )
        else:
            try:
                val = str(float(filter.value))
            except ValueError:
                if str(filter.value).lower() in [
                    "true",
                    "false",
                ]:
                    val = "True" if str(filter.value).lower() == "true" else "False"
                else:
                    val = '"{}"'.format(filter.value)
            return "(`{}` {} {}) {}".format(
                filter.column, filter.operation, val, filter.join
            )
    return ""


def filter_table(
    df,
    filter_predicates: List[Union[FilterPredicate, FilterPredicateGroup]],
    filter_ids: Optional[FilterIds] = None,
) -> pd.DataFrame:
    # if we have ids, filter them out now!
    if filter_ids is not None and len(filter_ids.ids) > 0:
        # this is fast because the index is set to ids
        df = df.loc[filter_ids.ids]

    final_filter = ""
    for filt in filter_predicates:
        final_filter = final_filter + get_filter_string(filt)
    if (
        len(filter_predicates) > 0
        and len(final_filter) > 0
        and (final_filter[-1] == "&" or final_filter[-1] == "|")
    ):
        final_filter = final_filter[0:-1]
    if len(final_filter) > 0:
        return df.query(final_filter)
    else:
        return df


def filter_table_single(df: DataFrame, col: ZenoColumn, bucket: HistogramBucket):
    if (
        col.metadata_type == MetadataType.NOMINAL
        or col.metadata_type == MetadataType.BOOLEAN
    ):
        return df[df[str(col)] == bucket.bucket]
    elif col.metadata_type == MetadataType.CONTINUOUS:
        return df[(df[str(col)] > bucket.bucket) & (df[str(col)] < bucket.bucket_end)]
    elif col.metadata_type == MetadataType.DATETIME:
        return df
    return df



"""Parallel processing functions for distill and inference pipelines."""

import os
from inspect import getsource
from pathlib import Path
from typing import Callable, Tuple

import pandas as pd
from tqdm import trange

from zeno.api import ZenoOptions
from zeno.classes.base import ZenoColumn, ZenoColumnType


def predistill_data(
    fn: Callable,
    column: ZenoColumn,
    options: ZenoOptions,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
) -> Tuple[ZenoColumn, pd.Series]:
    col_hash = str(column)
    col = df[col_hash]

    save_path = Path(cache_path, col_hash + ".pickle")
    to_predict_indices = col.loc[pd.isna(col)].index

    if len(to_predict_indices) > 0:
        if len(to_predict_indices) < batch_size:
            out = fn(df.loc[to_predict_indices], options)
            col.loc[to_predict_indices] = out
            col.to_pickle(str(save_path))
        else:
            for i in trange(
                0,
                len(to_predict_indices),
                batch_size,
                desc="preprocessing " + fn.__name__,
                position=pos,
            ):
                out = fn(df.loc[to_predict_indices[i : i + batch_size]], options)
                col.loc[to_predict_indices[i : i + batch_size]] = out
                col.to_pickle(str(save_path))
    return (column, col)


def run_inference(
    fn: Callable,
    options: ZenoOptions,
    model_path: str,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
) -> Tuple[ZenoColumn, ZenoColumn, pd.Series, pd.Series]:
    model_name = os.path.basename(model_path).split(".")[0]

    model_col_obj = ZenoColumn(
        column_type=ZenoColumnType.OUTPUT,
        name=model_name,
    )
    embedding_col_obj = ZenoColumn(
        column_type=ZenoColumnType.EMBEDDING,
        name=model_name,
    )
    model_hash = str(model_col_obj)
    embedding_hash = str(embedding_col_obj)
    model_col = df[model_hash]
    embedding_col = df[embedding_hash]

    model_save_path = Path(cache_path, model_hash + ".pickle")
    embedding_save_path = Path(cache_path, embedding_hash + ".pickle")

    to_predict_indices = model_col.loc[pd.isna(model_col)].index

    if len(to_predict_indices) > 0:
        model_fn = fn(model_path)
        if len(to_predict_indices) < batch_size:
            # Make output folder if function uses output_path
            src = getsource(model_fn)
            if "output_path" in src:
                file_cache_path = os.path.join(cache_path, model_hash)
                os.makedirs(file_cache_path, exist_ok=True)
                options = options.copy(update={"output_path": file_cache_path})

            out = model_fn(df.loc[to_predict_indices], options)

            # Check if we also get embedding
            if type(out) == tuple and len(out) == 2:
                for i, idx in enumerate(to_predict_indices):
                    model_col.at[idx] = out[0][i]
                    embedding_col.at[idx] = out[1][i]
                embedding_col.to_pickle(str(embedding_save_path))
                out = out[0]
            else:
                model_col[to_predict_indices] = out

            model_col.to_pickle(str(model_save_path))

        else:
            for i in trange(
                0,
                len(to_predict_indices),
                batch_size,
                desc="Inference on " + model_name,
                position=pos,
            ):

                # Make output folder if function uses output_path
                src = getsource(model_fn)
                if "output_path" in src:
                    file_cache_path = os.path.join(cache_path, model_hash)
                    os.makedirs(file_cache_path, exist_ok=True)
                    options = options.copy(update={"output_path": file_cache_path})

                out = model_fn(df.loc[to_predict_indices[i : i + batch_size]], options)

                # Check if we also get embedding
                if type(out) == tuple and len(out) == 2:
                    for i, idx in enumerate(to_predict_indices[i : i + batch_size]):
                        model_col.at[idx] = out[0][i]
                        embedding_col.at[idx] = out[1][i]
                    embedding_col.to_pickle(str(embedding_save_path))
                    out = out[0]
                else:
                    model_col.loc[to_predict_indices[i : i + batch_size]] = out

                model_col.to_pickle(str(model_save_path))

    return (model_col_obj, embedding_col_obj, model_col, embedding_col)


def postdistill_data(
    fn: Callable,
    model: str,
    options: ZenoOptions,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
):
    col_obj = ZenoColumn(
        column_type=ZenoColumnType.POSTDISTILL,
        name=fn.__name__,
        model=model,
    )
    col_hash = str(col_obj)
    col = df[col_hash]

    output_obj = ZenoColumn(
        column_type=ZenoColumnType.OUTPUT,
        name=model,
    )
    output_hash = str(output_obj)

    save_path = Path(cache_path, col_hash + ".pickle")

    to_predict_indices = col.loc[pd.isna(col)].index

    local_options = options.copy(
        update={
            "output_column": output_hash,
            "output_path": os.path.join(cache_path, output_hash),
        }
    )

    if len(to_predict_indices) > 0:
        if len(to_predict_indices) < batch_size:
            out = fn(df.loc[to_predict_indices], local_options)
            col.loc[to_predict_indices] = out
            col.to_pickle(str(save_path))
        else:
            for i in trange(
                0,
                len(to_predict_indices),
                batch_size,
                desc="postprocessing " + fn.__name__ + " on " + model,
                position=pos,
            ):
                out = fn(df.loc[to_predict_indices[i : i + batch_size]], local_options)
                col.loc[to_predict_indices[i : i + batch_size]] = out
                col.to_pickle(str(save_path))
    return (col_obj, col)



"""Primary backend for Zeno. Handles all data processing and caching."""

import asyncio
import logging
import os
import pickle
import sys
import threading
from inspect import getsource
from math import isnan
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from methodtools import lru_cache  # type: ignore
from pandas import DataFrame  # type: ignore
from pathos.multiprocessing import ProcessingPool as Pool  # type: ignore
from sklearn import preprocessing  # type: ignore

from zeno.api import ZenoOptions, ZenoParameters
from zeno.classes.base import MetadataType, ZenoColumnType
from zeno.classes.classes import MetricKey, TableRequest, ZenoColumn
from zeno.classes.metadata import HistogramBucket, HistogramRequest
from zeno.classes.projection import Points2D
from zeno.classes.report import Report
from zeno.classes.slice import (
    FilterIds,
    FilterPredicate,
    FilterPredicateGroup,
    Slice,
    SliceMetric,
)
from zeno.data_pipeline.data_processing import (
    postdistill_data,
    predistill_data,
    run_inference,
)
from zeno.data_pipeline.filtering import filter_table, filter_table_single
from zeno.util import getMetadataType, load_series, read_pickle


class ZenoBackend(object):
    def __init__(self, params: ZenoParameters):
        logging.basicConfig(level=logging.INFO)

        self.df = params.metadata
        self.tests = params.functions
        self.batch_size = params.batch_size
        self.data_path = params.data_path
        self.label_path = params.label_path
        self.cache_path = params.cache_path
        self.editable = params.editable
        self.samples = params.samples
        self.view = params.view

        self.done_running_inference = False

        self.predistill_functions: Dict[str, Callable] = {}
        self.postdistill_functions: Dict[str, Callable] = {}
        self.metric_functions: Dict[str, Callable] = {}
        self.predict_function: Optional[Callable] = None

        self.status: str = "Initializing"
        self.folders: List[str] = read_pickle("folders.pickle", self.cache_path, [])
        self.reports: List[Report] = read_pickle("reports.pickle", self.cache_path, [])
        self.slices: Dict[str, Slice] = read_pickle(
            "slices.pickle", self.cache_path, {}
        )

        if params.models and os.path.isdir(params.models[0]):
            self.model_paths = [
                os.path.join(params.models[0], m) for m in os.listdir(params.models[0])
            ]
        else:
            self.model_paths = params.models  # type: ignore
        self.model_names = [os.path.basename(p).split(".")[0] for p in self.model_paths]

        self.__setup_dataframe(
            params.id_column, params.data_column, params.label_column
        )
        self.__parse_test_functions(self.tests)

        # Options passed to Zeno functions.
        self.zeno_options = ZenoOptions(
            id_column=str(self.id_column),
            data_column=str(self.data_column),
            label_column=str(self.label_column),
            distill_columns=dict(),
            data_path=self.data_path,
            label_path=self.label_path,
            output_column="",
            output_path="",
        )

    def __setup_dataframe(self, id_column: str, data_column: str, label_column: str):
        if id_column != "":
            self.id_column = ZenoColumn(
                column_type=ZenoColumnType.METADATA,
                metadata_type=getMetadataType(self.df[id_column]),
                name=id_column,
            )
        else:
            self.df.reset_index(inplace=True)
            self.id_column = ZenoColumn(
                column_type=ZenoColumnType.METADATA,
                metadata_type=MetadataType.OTHER,
                name="index",
            )

        if data_column != "":
            self.data_column = ZenoColumn(
                column_type=ZenoColumnType.METADATA,
                metadata_type=getMetadataType(self.df[data_column]),
                name=data_column,
            )
        else:
            self.data_column = ZenoColumn(
                column_type=ZenoColumnType.METADATA,
                metadata_type=MetadataType.OTHER,
                name="",
            )

        if label_column != "":
            self.label_column = ZenoColumn(
                column_type=ZenoColumnType.METADATA,
                metadata_type=getMetadataType(self.df[label_column]),
                name=label_column,
            )
        else:
            self.label_column = ZenoColumn(
                column_type=ZenoColumnType.METADATA,
                metadata_type=MetadataType.OTHER,
                name="",
            )

        self.df[str(self.id_column)].astype(str)
        self.df.set_index(str(self.id_column), inplace=True)
        self.df[str(self.id_column)] = self.df.index

        self.columns: List[ZenoColumn] = []
        self.complete_columns: List[ZenoColumn] = []
        for metadata_col in self.df.columns:
            col = ZenoColumn(
                column_type=ZenoColumnType.METADATA,
                metadata_type=getMetadataType(self.df[metadata_col]),
                name=str(metadata_col),
            )
            self.columns.append(col)
            self.complete_columns.append(col)

    def __parse_test_functions(self, tests: List[Callable]):
        for test_fn in tests:
            if hasattr(test_fn, "predict_function"):
                if self.predict_function is None:
                    self.predict_function = test_fn
                else:
                    print("ERROR: Multiple model functions found, can only have one")
                    sys.exit(1)
            if hasattr(test_fn, "distill_function"):
                src = getsource(test_fn)
                if "output_column" in src:
                    self.postdistill_functions[test_fn.__name__] = test_fn
                else:
                    self.predistill_functions[test_fn.__name__] = test_fn
            if hasattr(test_fn, "metric_function"):
                self.metric_functions[test_fn.__name__] = test_fn

    def start_processing(self):
        """Parse testing files, distill, and run inference."""

        if not self.tests:
            self.done_running_inference = True
            self.status = "Done processing"
            return

        for fn in self.predistill_functions.values():
            self.columns.append(
                ZenoColumn(column_type=ZenoColumnType.PREDISTILL, name=fn.__name__)
            )
        for fn in self.postdistill_functions.values():
            for m in self.model_names:
                self.columns.append(
                    ZenoColumn(
                        column_type=ZenoColumnType.POSTDISTILL,
                        name=fn.__name__,
                        model=m,
                    )
                )

        self.__thread = threading.Thread(target=asyncio.run, args=(self.__process(),))
        self.__thread.start()

    async def __process(self):
        self.status = "Running predistill functions"
        print(self.status)
        self.__predistill()

        self.status = "Running inference"
        print(self.status)
        self.__inference()
        self.done_running_inference = True

        self.status = "Running postdistill functions"
        print(self.status)
        self.__postdistill()

        self.status = "Done processing"
        print(self.status)

    def __predistill(self) -> None:
        """Run distilling functions not dependent on model outputs."""

        # Check if we need to preprocess since Pool is expensive
        predistill_to_run: List[ZenoColumn] = []
        for predistill_column in [
            c for c in self.columns if c.column_type == ZenoColumnType.PREDISTILL
        ]:
            save_path = Path(self.cache_path, str(predistill_column) + ".pickle")

            load_series(self.df, predistill_column, save_path)

            if self.df[str(predistill_column)].isnull().any():
                predistill_to_run.append(predistill_column)
            else:
                predistill_column.metadata_type = getMetadataType(
                    self.df[str(predistill_column)]
                )
                self.complete_columns.append(predistill_column)

        if len(predistill_to_run) > 0:
            with Pool() as pool:
                predistill_outputs = pool.map(
                    predistill_data,
                    [self.predistill_functions[col.name] for col in predistill_to_run],
                    [col for col in predistill_to_run],
                    [self.zeno_options] * len(predistill_to_run),
                    [self.cache_path] * len(predistill_to_run),
                    [self.df] * len(predistill_to_run),
                    [self.batch_size] * len(predistill_to_run),
                    range(len(predistill_to_run)),
                )
                for out in predistill_outputs:
                    out[0].metadata_type = getMetadataType(out[1])
                    self.df.loc[:, str(out[0])] = out[1]
                    self.complete_columns.append(out[0])

    def __inference(self):
        """Run models on instances."""

        # Check if we need to run inference since Pool is expensive
        models_to_run = []
        for model_path in self.model_paths:
            model_name = os.path.basename(model_path).split(".")[0]
            model_column = ZenoColumn(
                column_type=ZenoColumnType.OUTPUT,
                name=model_name,
            )
            embedding_column = ZenoColumn(
                column_type=ZenoColumnType.EMBEDDING,
                name=model_name,
            )
            model_hash = str(model_column)
            embedding_hash = str(embedding_column)

            model_save_path = Path(self.cache_path, model_hash + ".pickle")
            embedding_save_path = Path(self.cache_path, embedding_hash + ".pickle")

            load_series(self.df, model_column, model_save_path)
            load_series(self.df, embedding_column, embedding_save_path)

            if (
                self.df[model_hash].isnull().any()
                or self.df[embedding_hash].isnull().any()
            ):
                models_to_run.append(model_path)
            else:
                self.complete_columns.append(model_column)

        if len(models_to_run) > 0:
            with Pool() as pool:
                inference_outputs = pool.map(
                    run_inference,
                    [self.predict_function] * len(models_to_run),
                    [self.zeno_options] * len(models_to_run),
                    [m for m in models_to_run],
                    [self.cache_path] * len(models_to_run),
                    [self.df] * len(models_to_run),
                    [self.batch_size] * len(models_to_run),
                    range(len(models_to_run)),
                )
                for out in inference_outputs:
                    self.df.loc[:, str(out[0])] = out[2]
                    if not out[3].isnull().values.any():  # type: ignore
                        self.df.loc[:, str(out[1])] = out[3]
                    self.complete_columns.append(out[0])

    def __postdistill(self) -> None:
        """Run distill functions dependent on model outputs."""

        # Check if we need to run postprocessing since Pool is expensive
        postdistill_to_run: List[ZenoColumn] = []
        for postdistill_column in [
            c for c in self.columns if c.column_type == ZenoColumnType.POSTDISTILL
        ]:
            col_name = postdistill_column.copy(
                update={
                    "model": postdistill_column.model,
                }
            )
            col_hash = str(col_name)
            save_path = Path(self.cache_path, col_hash + ".pickle")

            load_series(self.df, col_name, save_path)

            if self.df[col_hash].isnull().any():
                postdistill_to_run.append(col_name)
            else:
                col_name.metadata_type = getMetadataType(self.df[col_hash])
                self.complete_columns.append(col_name)

        if len(postdistill_to_run) > 0:
            with Pool() as pool:
                post_outputs = pool.map(
                    postdistill_data,
                    [self.postdistill_functions[e.name] for e in postdistill_to_run],
                    [e.model for e in postdistill_to_run],
                    [self.zeno_options] * len(postdistill_to_run),
                    [self.cache_path] * len(postdistill_to_run),
                    [self.df] * len(postdistill_to_run),
                    [self.batch_size] * len(postdistill_to_run),
                    range(len(postdistill_to_run)),
                )
                for out in post_outputs:
                    out[0].metadata_type = getMetadataType(out[1])
                    self.df.loc[:, str(out[0])] = out[1]  # type: ignore
                    self.complete_columns.append(out[0])

    def get_metrics_for_slices(
        self,
        requests: List[MetricKey],
        filter_ids: Optional[FilterIds] = None,
    ) -> List[SliceMetric]:
        """Calculate result for each requested combination."""

        return_metrics: List[SliceMetric] = []
        for metric_key in requests:
            filt_df = filter_table(
                self.df, [metric_key.sli.filter_predicates], filter_ids
            )
            if metric_key.metric == "" or self.label_column.name == "":
                return_metrics.append(SliceMetric(metric=None, size=filt_df.shape[0]))
            else:
                metric = self.calculate_metric(
                    filt_df, metric_key.model, metric_key.metric
                )
                return_metrics.append(SliceMetric(metric=metric, size=filt_df.shape[0]))
        return return_metrics

    def calculate_metric(
        self, df: DataFrame, model: Union[str, None], metric: str
    ) -> Union[float, None]:
        if not self.done_running_inference:
            return None

        if model is not None:

            output_col = ZenoColumn(
                column_type=ZenoColumnType.OUTPUT,
                name=model,
            )
            output_hash = str(output_col)

            distill_fns = [
                c
                for c in self.columns
                if (
                    c.column_type == ZenoColumnType.PREDISTILL
                    or c.column_type == ZenoColumnType.POSTDISTILL
                )
                and c.model == model
            ]

            local_ops = self.zeno_options.copy(
                update={
                    "output_column": output_hash,
                    "output_path": os.path.join(self.cache_path, output_hash),
                    "distill_columns": dict(
                        zip(
                            [c.name for c in distill_fns], [str(c) for c in distill_fns]
                        )
                    ),
                }
            )
        else:
            distill_fns = [
                c
                for c in self.columns
                if (
                    c.column_type == ZenoColumnType.PREDISTILL
                    or c.column_type == ZenoColumnType.POSTDISTILL
                )
            ]

            local_ops = self.zeno_options.copy(
                update={
                    "distill_columns": dict(
                        zip(
                            [c.name for c in distill_fns], [str(c) for c in distill_fns]
                        )
                    ),
                }
            )

        return self.metric_functions[metric](df, local_ops)

    def set_folders(self, folders: List[str]):
        if not self.editable:
            return
        self.folders = folders
        with open(os.path.join(self.cache_path, "folders.pickle"), "wb") as f:
            pickle.dump(self.folders, f)

    def set_reports(self, reports: List[Report]):
        if not self.editable:
            return
        self.reports = reports
        with open(os.path.join(self.cache_path, "reports.pickle"), "wb") as f:
            pickle.dump(self.reports, f)

    def create_new_slice(self, req: Slice):
        if not self.editable:
            return
        self.slices[req.slice_name] = req
        with open(os.path.join(self.cache_path, "slices.pickle"), "wb") as f:
            pickle.dump(self.slices, f)

    def delete_slice(self, slice_name: str):
        if not self.editable:
            return
        del self.slices[slice_name]
        with open(os.path.join(self.cache_path, "slices.pickle"), "wb") as f:
            pickle.dump(self.slices, f)

    def get_filtered_ids(self, req: List[Union[FilterPredicateGroup, FilterPredicate]]):
        return filter_table(self.df, req)[str(self.id_column)].to_json(orient="records")

    def get_filtered_table(self, req: TableRequest):
        """Return filtered table from list of filter predicates."""
        filt_df = filter_table(self.df, req.filter_predicates, req.filter_ids)
        if req.sort[0]:
            filt_df = filt_df.sort_values(str(req.sort[0]), ascending=req.sort[1])
        filt_df = filt_df.iloc[req.slice_range[0] : req.slice_range[1]]
        return filt_df[[str(col) for col in req.columns]].to_json(orient="records")

    def get_histogram_buckets(
        self, req: List[ZenoColumn]
    ) -> List[List[HistogramBucket]]:
        """Calculate the histogram buckets for a list of columns."""
        res: List[List[HistogramBucket]] = []
        for col in req:
            df_col = self.df[str(col)]
            if col.metadata_type == MetadataType.NOMINAL:
                ret_hist: List[HistogramBucket] = []
                val_counts = df_col.value_counts()
                for k in val_counts.keys():
                    ret_hist.append(HistogramBucket(bucket=k))
                res.append(ret_hist)
            elif col.metadata_type == MetadataType.CONTINUOUS:
                ret_hist: List[HistogramBucket] = []  # type: ignore
                df_col = df_col.fillna(0)
                bins = np.histogram(df_col, bins="doane")
                for i in range(len(bins[0])):
                    ret_hist.append(
                        HistogramBucket(
                            bucket=round(bins[1][i], 4),
                            bucket_end=round(bins[1][i + 1], 4),
                        )
                    )
                res.append(ret_hist)
            elif col.metadata_type == MetadataType.BOOLEAN:
                res.append(
                    [
                        HistogramBucket(bucket=True),
                        HistogramBucket(bucket=False),
                    ]
                )
            elif col.metadata_type == MetadataType.DATETIME:
                res.append([])
            else:
                res.append([])
        return res

    def get_histogram_counts(self, req: HistogramRequest) -> List[List[int]]:
        """Calculate count for each bucket in each column histogram."""
        if req.filter_predicates is not None:
            filt_df = filter_table(self.df, [req.filter_predicates], req.filter_ids)
        else:
            filt_df = self.df

        ret: List[List[int]] = []
        for r in req.column_requests:
            col = r.column
            ret.append([len(filter_table_single(filt_df, col, b)) for b in r.buckets])
        return ret

    def get_histogram_metric(
        self,
        df: pd.DataFrame,
        col: ZenoColumn,
        bucket: HistogramBucket,
        model: str,
        metric: str,
    ) -> Union[float, None]:
        df_filt = filter_table_single(df, col, bucket)
        output_metric = self.calculate_metric(df_filt, model, metric)
        if output_metric is None or isnan(output_metric):
            return None
        return output_metric

    def get_histogram_metrics(
        self, req: HistogramRequest
    ) -> List[List[Union[float, None]]]:
        """Calculate metric for each bucket in each column histogram."""
        if req.metric is None:
            return []

        if req.filter_predicates is not None:
            filt_df = filter_table(self.df, [req.filter_predicates], req.filter_ids)
        else:
            filt_df = self.df

        ret: List[List[Union[float, None]]] = []
        for r in req.column_requests:
            col = r.column
            loc_ret: List[Union[float, None]] = []
            for b in r.buckets:
                df_filt = filter_table_single(filt_df, col, b)
                metric = self.calculate_metric(df_filt, req.model, req.metric)
                if metric is None or isnan(metric):
                    loc_ret.append(None)
                else:
                    loc_ret.append(metric)
            ret.append(loc_ret)
        return ret

    def embed_exists(self, model: str):
        """Checks for the existence of an embedding column.
        Returns True if the column exists, False otherwise
        """
        embed_column = ZenoColumn(name=model, column_type=ZenoColumnType.EMBEDDING)
        exists = str(embed_column) in self.df.columns
        return exists and not self.df[str(embed_column)].isnull().any()

    @lru_cache(
        maxsize=5
    )  # will cache up 5 model projections in memory before needing to recompute
    def run_tsne(self, model: str) -> np.ndarray:
        # Extract embeddings and store in one big ndarray
        embed_col = ZenoColumn(column_type=ZenoColumnType.EMBEDDING, name=model)

        embed = self.df[str(embed_col)].to_numpy()
        embed = np.stack(embed, axis=0)  # type: ignore

        # project embeddings into 2D
        from sklearn.manifold import TSNE  # type: ignore

        return TSNE().fit_transform(embed)

    def get_projection_colors(self, column: ZenoColumn) -> Tuple[List[int], List, str]:
        """Get colors for a projection based on a column.

        Args:
            column (ZenoColumn): Column to use for coloring the projection.

        Returns:
            Tuple[List[int], List, str]: The color range, the unique values,
                and the metadata type.
        """
        series = self.df[str(column)]
        unique = series.unique()
        metadata_type = "nominal"
        color_range: List[int] = []
        if len(unique) == 2:
            metadata_type = "boolean"
        if len(unique) > 10:
            if column.metadata_type == MetadataType.CONTINUOUS:
                metadata_type = "continuous"
                color_range = (
                    np.interp(series, (series.min(), series.max()), (0, 20))
                    .astype(int)
                    .tolist()
                )
                unique = np.array([series.min(), series.max()])
            else:
                metadata_type = "other"
                color_range = [0] * len(series)
        else:
            labels = preprocessing.LabelEncoder().fit_transform(series)
            if isinstance(labels, np.ndarray):
                color_range = labels.astype(int).tolist()
            else:
                color_range = [0] * len(series)
        return color_range, unique.tolist(), metadata_type

    def project_embed_into_2D(self, model: str, column: ZenoColumn) -> Points2D:
        """If the embedding exists, will use t-SNE to project into 2D."""

        points = Points2D(
            x=[], y=[], color=[], domain=[], opacity=[], dataType="", ids=[]
        )

        # Can't project without an embedding
        if not self.embed_exists(model):
            return points

        projection = self.run_tsne(model)

        # extract points and ids from computed projection
        points.x = projection[:, 0].tolist()
        points.y = projection[:, 1].tolist()
        color_results = self.get_projection_colors(column)
        points.color = color_results[0]
        points.domain = color_results[1]
        points.dataType = color_results[2]
        points.ids = self.df[str(self.id_column)].to_list()

        return points



"""Entry to Zeno. Parses TOML file, starts server, and runs the pipeline."""

import os
import shutil
import sys
from pathlib import Path
from typing import Union

import pandas as pd
import pkg_resources
import requests
import tomli
import uvicorn
from multiprocess import Process  # type: ignore

from zeno.api import ZenoParameters
from zeno.data_pipeline.zeno_backend import ZenoBackend
from zeno.server import get_server
from zeno.setup import setup_zeno
from zeno.util import is_notebook, parse_testing_file, VIEW_MAP_URL, VIEWS_MAP_JSON


# Global variable to hold the Zeno server process.
# This is used to kill the server when re-running in a notebook.
ZENO_SERVER_PROCESS = None


def command_line():
    if len(sys.argv) == 1 or sys.argv[1] == "-h" or sys.argv[1] == "--help":
        print(
            "\n \033[1mZeno\033[0m",
            pkg_resources.get_distribution("zenoml").version,
            " - Machine learning evaluation framework.",
            "\n\n",
            "\033[1mUSAGE \033[0m \n\t",
            "zeno [-h] [-v] <config.toml>",
            "\n\n",
            "\033[1mARGUMENTS \033[0m \n\t",
            "<config.toml>\t\tZeno configuration file.\n\n"
            "\033[1m GLOBAL OPTIONS \033[0m \n\t",
            "-h (--help)\t\tDisplay this help message.\n"
            "\t -v (--version)\t\tDisplay this application version.\n",
        )

        sys.exit(0)

    if len(sys.argv) != 2:
        print(
            "ERROR: Zeno take one argument, either a configuration TOML file"
            + " or the keyword 'init'. "
            + "{0} arguments were passed.",
            len(sys.argv),
        )
        sys.exit(1)

    if sys.argv[1] == "-v" or sys.argv[1] == "--version":
        print(pkg_resources.get_distribution("zenoml").version)
        sys.exit(0)

    if sys.argv[1] == "init" or sys.argv[1] == "i":
        setup_zeno()
    else:
        parse_toml()


def parse_toml():
    args = {}
    try:
        with open(sys.argv[1], "rb") as f:
            args = tomli.load(f)
    except Exception:
        print("ERROR: Failed to read TOML configuration file.")
        sys.exit(1)

    base_path = os.path.dirname(os.path.abspath(sys.argv[1]))

    if "metadata" not in args:
        print("ERROR: Must have 'metadata' entry which must be a CSV or Parquet file.")
        sys.exit(1)
    else:
        meta_path = Path(os.path.realpath(os.path.join(base_path, args["metadata"])))

        # Read metadata as Pandas for slicing
        if meta_path.suffix == ".csv":
            args["metadata"] = pd.read_csv(meta_path)
        elif meta_path.suffix == ".parquet":
            args["metadata"] = pd.read_parquet(meta_path)
        else:
            print("Extension of " + meta_path.suffix + " not one of .csv or .parquet")
            sys.exit(1)

    if "functions" in args:
        fn_path = Path(os.path.realpath(os.path.join(base_path, args["functions"])))
        if os.path.isfile(fn_path):
            args["functions"] = parse_testing_file(fn_path)
        elif os.path.exists(fn_path):
            # Add directory with tests to path for relative imports.
            fns = []
            for f in list(fn_path.rglob("*.py")):
                fns = fns + parse_testing_file(f)
            args["functions"] = fns

    zeno(ZenoParameters(**args), base_path)


def parse_args(args: ZenoParameters, base_path) -> ZenoParameters:

    if type(args) == dict:
        args = ZenoParameters.parse_obj(args)

    if args.cache_path == "":
        args.cache_path = os.path.realpath(os.path.join(base_path, "./.zeno_cache/"))
    else:
        args.cache_path = os.path.realpath(os.path.join(base_path, args.cache_path))

    os.makedirs(args.cache_path, exist_ok=True)

    # Try to get view from GitHub List, if not try to read from path and copy it.
    if args.view != "":
        view_dest_path = Path(os.path.join(args.cache_path, "view.mjs"))
        try:
            views_res = requests.get(VIEW_MAP_URL + VIEWS_MAP_JSON)
            views = views_res.json()
            url = VIEW_MAP_URL + views[args.view]
            with open(view_dest_path, "wb") as out_file:
                content = requests.get(url, stream=True).content
                out_file.write(content)
        except requests.exceptions.ConnectionError:
            view_path = Path(os.path.realpath(os.path.join(base_path, args.view)))
            if view_path.is_file():
                if view_dest_path.is_file():
                    os.remove(view_dest_path)
                shutil.copyfile(view_path, view_dest_path)
        except KeyError:
            print(
                "ERROR: View not found in list or relative path."
                " See available views at ",
                "https://github.com/zeno-ml/instance-views/blob/main/views.json",
            )
            sys.exit(1)

    if len(args.models) > 0:
        if Path(os.path.realpath(os.path.join(base_path, args.models[0]))).exists():
            args.models = [
                os.path.realpath(os.path.join(base_path, m)) for m in args.models
            ]

    if not args.data_path.startswith("http"):
        args.data_path = os.path.realpath(os.path.join(base_path, args.data_path))

    if args.label_path != "":
        args.label_path = os.path.realpath(os.path.join(base_path, args.label_path))

    if args.data_column != "" and args.id_column == "":
        args.id_column = args.data_column

    return args


def run_zeno(args: ZenoParameters):
    zeno = ZenoBackend(args)
    app = get_server(zeno)
    zeno.start_processing()

    print("\n\033[1mZeno\033[0m running on http://{}:{}\n".format(args.host, args.port))
    uvicorn.run(app, host=args.host, port=args.port, log_level="error")


def zeno(args: Union[ZenoParameters, dict], base_path="./"):
    if isinstance(args, dict):
        args = ZenoParameters.parse_obj(args)
    args = parse_args(args, base_path)

    if args.serve:
        global ZENO_SERVER_PROCESS
        if ZENO_SERVER_PROCESS is not None:
            ZENO_SERVER_PROCESS.terminate()

        ZENO_SERVER_PROCESS = Process(target=run_zeno, args=(args,))
        ZENO_SERVER_PROCESS.start()

        if not is_notebook():
            ZENO_SERVER_PROCESS.join()

    else:
        zeno = ZenoBackend(args)
        return zeno



"""
The FastAPI server for the Zeno frontend. Hosts both the REST API
and the static files for the frontend.
"""

import asyncio
import os
from typing import Dict, List, Union

from fastapi import FastAPI, HTTPException, WebSocket
from fastapi.routing import APIRoute
from fastapi.staticfiles import StaticFiles

from zeno.classes.base import ZenoColumn
from zeno.classes.classes import (
    EmbedProject2DRequest,
    EntryRequest,
    MetricRequest,
    StatusResponse,
    TableRequest,
    ZenoSettings,
    ZenoVariables,
)
from zeno.classes.metadata import HistogramBucket, HistogramRequest
from zeno.classes.projection import Points2D
from zeno.classes.report import Report
from zeno.classes.slice import FilterPredicate, FilterPredicateGroup, Slice, SliceMetric
from zeno.data_pipeline.zeno_backend import ZenoBackend


def custom_generate_unique_id(route: APIRoute):
    return route.name


def get_server(zeno: ZenoBackend):
    app = FastAPI(title="Frontend API")
    api_app = FastAPI(
        title="Backend API", generate_unique_id_function=custom_generate_unique_id
    )

    if zeno.data_path != "" and os.path.exists(zeno.data_path):
        app.mount("/data", StaticFiles(directory=zeno.data_path), name="static")
    if zeno.label_path != "" and os.path.exists(zeno.label_path):
        app.mount("/labels", StaticFiles(directory=zeno.label_path), name="static")

    app.mount(
        "/cache",
        StaticFiles(directory=zeno.cache_path),
        name="cache",
    )
    app.mount("/api", api_app)
    app.mount(
        "/",
        StaticFiles(
            directory=os.path.dirname(os.path.realpath(__file__)) + "/frontend",
            html=True,
        ),
        name="base",
    )

    @api_app.get("/settings", response_model=ZenoSettings, tags=["zeno"])
    def get_settings():
        return ZenoSettings(
            view=zeno.view,
            id_column=zeno.id_column,
            label_column=zeno.label_column,
            data_column=zeno.data_column,
            data_origin="/data/" if os.path.exists(zeno.data_path) else zeno.data_path,
            metadata_columns=zeno.columns,
            samples=zeno.samples,
            totalSize=zeno.df.shape[0],
        )

    @api_app.get("/initialize", response_model=ZenoVariables, tags=["zeno"])
    def get_initial_info():
        return ZenoVariables(
            metrics=list(zeno.metric_functions.keys()),
            models=[str(n) for n in zeno.model_names],
            folders=zeno.folders,
        )

    @api_app.get("/slices", response_model=Dict[str, Slice], tags=["zeno"])
    def get_slices():
        return zeno.slices

    @api_app.get("/reports", response_model=List[Report], tags=["zeno"])
    def get_reports():
        return zeno.reports

    @api_app.post("/folders", tags=["zeno"])
    def set_folders(folders: List[str]):
        zeno.set_folders(folders)

    @api_app.post("/reports", tags=["zeno"])
    def update_reports(reqs: List[Report]):
        zeno.set_reports(reqs)

    @api_app.post("/filtered-ids", response_model=str, tags=["zeno"])
    def get_filtered_ids(req: List[Union[FilterPredicateGroup, FilterPredicate]]):
        return zeno.get_filtered_ids(req)

    @api_app.post("/filtered-table", response_model=str, tags=["zeno"])
    def get_filtered_table(req: TableRequest):
        return zeno.get_filtered_table(req)

    @api_app.post(
        "/histograms", response_model=List[List[HistogramBucket]], tags=["zeno"]
    )
    def get_histogram_buckets(req: List[ZenoColumn]):
        return zeno.get_histogram_buckets(req)

    @api_app.post("/histogram-counts", response_model=List[List[int]], tags=["zeno"])
    def calculate_histogram_counts(req: HistogramRequest):
        return zeno.get_histogram_counts(req)

    @api_app.post(
        "/histogram-metrics",
        response_model=List[List[Union[float, None]]],
        tags=["zeno"],
    )
    def calculate_histogram_metrics(req: HistogramRequest):
        return zeno.get_histogram_metrics(req)

    @api_app.post("/slice", tags=["zeno"])
    def create_new_slice(req: Slice):
        zeno.create_new_slice(req)

    @api_app.delete("/slice", tags=["zeno"])
    def delete_slice(slice_name: List[str]):
        zeno.delete_slice(slice_name[0])

    @api_app.post("/slice-metrics", response_model=List[SliceMetric], tags=["zeno"])
    def get_metrics_for_slices(req: MetricRequest):
        return zeno.get_metrics_for_slices(req.metric_keys, req.filter_ids)

    @api_app.get("/embed-exists/{model}", response_model=bool, tags=["zeno"])
    def embed_exists(model: str):
        """Checks if embedding exists for a model.
        Returns the boolean True or False directly
        """
        return zeno.embed_exists(model)

    @api_app.post("/embed-project", tags=["zeno"], response_model=Points2D)
    def project_embed_into_2D(req: EmbedProject2DRequest):
        return zeno.project_embed_into_2D(req.model, req.column)

    @api_app.post("/entry", tags=["zeno"], response_model=str)
    def get_df_row_entry(req: EntryRequest):
        try:
            entry = zeno.df.loc[req.id, :]
            if len(req.columns) > 0:
                entry = entry[list(map(str, req.columns))]
            json_entry = entry.to_json()
            return json_entry
        except KeyError:
            raise HTTPException(
                status_code=404, detail=f"Entry with id={req.id} not found"
            )

    @api_app.websocket("/status")
    async def results_websocket(websocket: WebSocket):
        await websocket.accept()
        previous_status = ""
        while True:
            await asyncio.sleep(1)
            if zeno.status != previous_status:
                previous_status = zeno.status
                await websocket.send_json(
                    StatusResponse(
                        status=zeno.status,
                        done_processing=zeno.done_running_inference,
                        complete_columns=zeno.complete_columns,
                    ).json(by_alias=True)
                )

    return app



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zeno', 'zeno.classes', 'zeno.data_pipeline']

package_data = \
{'': ['*'],
 'zeno': ['frontend/*', 'frontend/build/*', 'frontend/build/assets/*']}

install_requires = \
['fastapi>=0.75,<0.90',
 'inquirerpy>=0.3.4,<0.4.0',
 'methodtools>=0.4.5,<0.5.0',
 'nest-asyncio>=1.5.6,<2.0.0',
 'pandas>=1.4.0,<2.0.0',
 'pathos>=0.3.0,<0.4.0',
 'requests>=2.28.1,<3.0.0',
 'scikit-learn>=1.1.2,<2.0.0',
 'setuptools>=67.0.0,<68.0.0',
 'tomli>=2.0.1,<3.0.0',
 'tqdm>=4.64.0,<5.0.0',
 'types-requests>=2.28.11,<3.0.0',
 'types-setuptools>=65.4.0,<66.0.0',
 'types-tqdm>=4.64.7,<5.0.0',
 'uvicorn>=0.17.5,<0.21.0',
 'websockets>=10.2,<11.0']

entry_points = \
{'console_scripts': ['zeno = zeno.runner:command_line']}

setup_kwargs = {
    'name': 'zenoml',
    'version': '0.3.2',
    'description': 'Interactive Evaluation Framework for Machine Learning',
    'long_description': '<img src="./zeno.png" width="250px"/>\n\n[![PyPI version](https://badge.fury.io/py/zenoml.svg)](https://badge.fury.io/py/zenoml)\n![Github Actions CI tests](https://github.com/zeno-ml/zeno/actions/workflows/test.yml/badge.svg)\n![Github Actions Docs build](https://github.com/zeno-ml/zenoml.com/actions/workflows/docs.yml/badge.svg)\n[![MIT license](https://img.shields.io/badge/License-MIT-blue.svg)](https://lbesson.mit-license.org/) [![Join the chat at https://gitter.im/zeno-ml-eval/community](https://badges.gitter.im/zeno-ml-eval/community.svg)](https://gitter.im/zeno-ml-eval/community?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge)\n\nZeno is a general-purpose framework for evaluating machine learning models.\nIt combines a **Python API** with an **interactive UI** to allow users to discover, explore, and analyze the performance of their models across diverse use cases.\nZeno can be used for any data type or task with [modular views](https://zenoml.com/docs/views/) for everything from object detection to audio transcription.\n\n## Quickstart\n\nInstall the Zeno Python package from PyPI:\n\n```bash\npip install zenoml\n```\n\n### Command Line\n\nTo get started, run the following command to initialize a Zeno project. It will walk you through creating the `zeno.toml` configuration file:\n\n```bash\nzeno init\n```\n\nThen run `zeno zeno.toml`.\n\n### Jupyter Notebook\n\nYou can also run Zeno directly from Jupyter notebooks or lab. The `zeno` command takes a dictionary of configuration options as input. See [the docs](https://zenoml.com/docs/configuration) for a full list of options. In this example we pass the minimum options for exploring a non-tabular dataset:\n\n```python\nimport pandas as pd\nfrom zeno import zeno\n\ndf = pd.read_csv("/path/to/metadata/file.csv")\n\nzeno({\n    "metadata": df, # Pandas DataFrame with a row for each instance\n    "view": "audio-transcription", # The type of view for this data/task\n    "data_path": "/path/to/raw/data/", # The folder with raw data (images, audio, etc.)\n    "data_column": "id" # The column in the metadata file that contains the relative paths of files in data_path\n})\n\n```\n\n## Learn More\n\nCheck out examples and additional documentation:\n\n- [Introduction](https://zenoml.com/docs/intro) - Learn more about Zeno.\n- [Quickstart](http://zenoml.com/docs/quickstart) - Setup Zeno with your own data and models.\n- [Documentation & API](http://zenoml.com/docs/intro) - Full documentation and API reference.\n',
    'author': 'Ángel Alexander Cabrera',
    'author_email': 'alex.cabrera@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://zenoml.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<=3.11',
}


setup(**setup_kwargs)



import datetime
import os
import pickle
import sys
from contextlib import contextmanager
from importlib import util
from inspect import getmembers, isfunction
from pathlib import Path

import pandas as pd

from zeno.classes.base import MetadataType

VIEW_MAP_URL: str = "https://raw.githubusercontent.com/zeno-ml/instance-views/main/"
VIEWS_MAP_JSON: str = "views.json"


def read_pickle(file_name: str, cache_path: str, default):
    try:
        with open(os.path.join(cache_path, file_name), "rb") as f:
            return pickle.load(f)
    except FileNotFoundError:
        return default


def getMetadataType(col: pd.Series) -> MetadataType:
    col = col.infer_objects()

    try:
        datetime.datetime.fromisoformat(str(col[0]))
        return MetadataType.DATETIME
    except ValueError:
        pass

    if col.dtype == "bool":
        return MetadataType.BOOLEAN

    try:
        unique = col.unique().tolist()
    except TypeError:
        return MetadataType.OTHER

    if len(unique) < 21:
        return MetadataType.NOMINAL

    if col.dtype in [
        "int64",
        "int32",
        "int16",
        "int8",
        "uint64",
        "uint32",
        "uint16",
        "uint8",
        "float64",
        "float32",
        "float16",
        "float8",
    ]:
        return MetadataType.CONTINUOUS

    return MetadataType.OTHER


def load_series(df, col_name, save_path):
    try:
        series = pd.read_pickle(save_path)
        col_name.metadata_type = getMetadataType(series)
        df.loc[:, str(col_name)] = series
    except FileNotFoundError:
        df.loc[:, str(col_name)] = pd.Series([pd.NA] * df.shape[0], index=df.index)
    except EOFError:
        df.loc[:, str(col_name)] = pd.Series([pd.NA] * df.shape[0], index=df.index)


@contextmanager
def add_to_path(p):
    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, p)
    try:
        yield
    finally:
        sys.path = old_path


def parse_testing_file(test_file: Path):
    # To allow relative imports in test files,
    # add their directory to path temporarily.
    with add_to_path(os.path.dirname(os.path.abspath(test_file))):
        spec = util.spec_from_file_location(str(test_file), test_file)
        test_module = util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(test_module)  # type: ignore

    functions = []
    for func_name, func in getmembers(test_module):
        if isfunction(func):
            if (
                hasattr(func, "predict_function")
                or hasattr(func, "distill_function")
                or hasattr(func, "metric_function")
            ):
                functions.append(func)
    return functions


def is_notebook() -> bool:
    try:
        from IPython import get_ipython  # type: ignore

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except (NameError, ImportError):
        return False  # Probably standard Python interpreter
